@@ -1,8 +1,13 @@
+"""See the class docstring."""
+
 from enum import Enum
 
-from .Conversions import str_to_str_list, str_to_bool, str_to_bool_list, str_to_int, str_to_int_list, str_to_float, str_to_float_list, str_to_scalar_enum_helper, str_to_list_enum_helper
+from .Conversions import str_to_bool, str_to_int, str_to_int_or_bool, str_to_float, str_to_float_or_bool, str_to_scalar_enum_helper
+from .Conversions import str_to_str_list, str_to_bool_list, str_to_int_list, str_to_int_or_bool_list, str_to_float_list, str_to_float_or_bool_list, str_to_list_enum_helper
 from .Types import DataType
 from .Utilities import trim_str_response
+from .InstrumentErrors import RsInstrException
+from .ScpiEnums import ScpiEnum
 
 
 class ConverterFromScpiString:
@@ -17,37 +22,50 @@ class ConverterFromScpiString:
 	On the other side, the driver methods might want to set the whole argument value, because the result scpi string is a single argument response."""
 
 	def __init__(self, data_type: DataType, enum_type: Enum = None):
-		self.enum_type = enum_type
+		self.scpi_enum = None
 		self.data_type = data_type
 		self.element_type = self.data_type.element_type
 
 		if self.element_type == DataType.RawString:
 			self.converter = trim_str_response
 			self.list_converter = str_to_str_list
+
 		elif self.element_type == DataType.String:
 			self.converter = trim_str_response
 			self.list_converter = str_to_str_list
+
 		elif self.element_type == DataType.Boolean:
 			self.converter = str_to_bool
 			self.list_converter = str_to_bool_list
+
 		elif self.element_type == DataType.Integer:
 			self.converter = str_to_int
 			self.list_converter = str_to_int_list
+
+		elif self.element_type == DataType.IntegerExt:
+			self.converter = str_to_int_or_bool
+			self.list_converter = str_to_int_or_bool_list
+
 		elif self.element_type == DataType.Float:
 			self.converter = str_to_float
 			self.list_converter = str_to_float_list
-		elif self.element_type == DataType.Enum:
-			assert self.enum_type, f"For data type enum, you have to define the enum_type variable."
+
+		elif self.element_type == DataType.FloatExt:
+			self.converter = str_to_float_or_bool
+			self.list_converter = str_to_float_or_bool_list
+
+		elif self.element_type.is_scalar_enum:
+			assert enum_type, f"For data type enum, you have to define the enum_type variable."
 			# noinspection PyTypeChecker
-			self.enum_members = [x.name for x in self.enum_type]
+			self.scpi_enum = ScpiEnum(enum_type)
 		else:
-			raise Exception(f"Unsupported data type '{data_type}'")
+			raise RsInstrException(f"Unsupported data type '{data_type}'")
 
 	def get_one_element_value(self, scpi_string: str):
 		"""Returns single element !!! of the argument value converted from the SCPI string (single element)"""
 		assert isinstance(scpi_string, str), f"Input parameter scpi_string must be string. Actual parameter: {type(scpi_string)}, value: {scpi_string}"
-		if self.element_type is DataType.Enum:
-			return str_to_scalar_enum_helper(scpi_string, self.enum_type, self.enum_members)
+		if self.element_type.is_scalar_enum:
+			return str_to_scalar_enum_helper(scpi_string, self.scpi_enum, False, exc_if_not_found=self.element_type == DataType.Enum)
 		return self.converter(scpi_string)
 
 	def get_value(self, scpi_string: str):
@@ -57,5 +75,5 @@ class ConverterFromScpiString:
 
 		assert isinstance(scpi_string, str), f"Input parameter scpi_string must be string. Actual parameter: {type(scpi_string)}, value: {scpi_string}"
 		if self.element_type is DataType.Enum:
-			return str_to_list_enum_helper(scpi_string, self.enum_type, self.enum_members)
+			return str_to_list_enum_helper(scpi_string, self.scpi_enum, exc_if_not_found=self.element_type == DataType.Enum)
 		return self.list_converter(scpi_string)
