@@ -1,11 +1,11 @@
 """VISA communication interface for SCPI-based instrument remote control.
-	:version: 1.8.1.41
-	:copyright: 2020 by Rohde & Schwarz GMBH & Co. KG
-	:license: MIT, see LICENSE for more details.
+    :version: 1.40.0.88
+    :copyright: 2020 by Rohde & Schwarz GMBH & Co. KG
+    :license: MIT, see LICENSE for more details.
 """
 
 
-__version__ = '1.8.1.41'
+__version__ = '1.40.0.88'
 
 # Main class
 from RsInstrument.RsInstrument import RsInstrument
@@ -18,3 +18,6 @@ from RsInstrument.Internal.InstrumentErrors import RsInstrException, TimeoutExce
 
 # Callback Event Argument prototypes
 from RsInstrument.Internal.IoTransferEventArgs import IoTransferEventArgs
+
+# Logging Mode
+from RsInstrument.Internal.ScpiLogger import LoggingMode
